@@ -18,7 +18,8 @@ import atexit
 import docker
 from shlex import quote as shellquote
 from datetime import datetime
-from argparse import ArgumentParser, Action, SUPPRESS
+from argparse import ArgumentParser, Action, SUPPRESS, RawDescriptionHelpFormatter
+from contextlib import ExitStack
 import importlib_metadata
 from . import *
 from ._util import (
@@ -26,7 +27,7 @@ from ._util import (
     write_values_json,
     VERBOSE_LEVEL,
     NOTICE_LEVEL,
-    install_coloredlogs,
+    configure_logger,
     parse_byte_size,
     path_really_within,
 )
@@ -37,7 +38,6 @@ quant_warning = False
 
 def main(args=None):
     sys.setrecursionlimit(1_000_000)  # permit as much call stack depth as OS can give us
-    os.environ["COLUMNS"] = os.environ.get("COLUMNS", "100")
 
     parser = ArgumentParser("miniwdl")
     parser.add_argument(
@@ -56,7 +56,14 @@ def main(args=None):
     fill_common(fill_localize_subparser(subparsers))
 
     argcomplete.autocomplete(parser)
+
+    replace_COLUMNS = os.environ.get("COLUMNS", None)
+    os.environ["COLUMNS"] = "100"  # make help descriptions wider
     args = parser.parse_args(args if args is not None else sys.argv[1:])
+    if replace_COLUMNS is not None:
+        os.environ["COLUMNS"] = replace_COLUMNS
+    else:
+        del os.environ["COLUMNS"]
 
     try:
         if args.command == "check":
@@ -135,25 +142,46 @@ def fill_common(subparser, path=True):
 
 def fill_check_subparser(subparsers):
     check_parser = subparsers.add_parser(
-        "check", help="Load and typecheck a WDL document; show an outline with lint warnings"
+        "check",
+        description="Load and typecheck a WDL document, showing an outline with lint warnings.\n\n"
+        "Individual lint warnings can be suppressed by a WDL comment containing !WarningName on the\n"
+        "same line or the following line, e.g.:\n"
+        "    Int? foo = 42  # !UnnecessaryQuantifier\n"
+        "    Int bar = foo + 1\n"
+        "    # Lorem ipsum dolor sit (!OptionalCoercion)\n",
+        formatter_class=RawDescriptionHelpFormatter,
     )
     check_parser.add_argument(
         "uri", metavar="URI", type=str, nargs="+", help="WDL document filename/URI"
     )
+    check_parser.add_argument(
+        "--strict",
+        action="store_true",
+        help="exit with nonzero status code if any lint warnings are shown (in addition to syntax and type errors)",
+    )
+    check_parser.add_argument(
+        "--no-suppress",
+        dest="show_all",
+        action="store_true",
+        help="show lint warnings even if they have suppression comments",
+    )
     check_parser.add_argument(
         "--no-shellcheck",
         dest="shellcheck",
         action="store_false",
-        help="don't use shellcheck on task commands even if available, and suppress message if it isn't",
+        help="don't use shellcheck on task commands even if available, and suppress suggestion if it isn't",
     )
     return check_parser
 
 
-def check(uri=None, path=None, check_quant=True, shellcheck=True, **kwargs):
+def check(
+    uri=None, path=None, check_quant=True, shellcheck=True, strict=False, show_all=False, **kwargs
+):
     # Load the document (read, parse, and typecheck)
     if not shellcheck:
         Lint._shellcheck_available = False
 
+    shown = [0]
     for uri1 in uri or []:
         doc = load(uri1, path or [], check_quant=check_quant, read_source=read_source)
 
@@ -161,16 +189,19 @@ def check(uri=None, path=None, check_quant=True, shellcheck=True, **kwargs):
 
         # Print an outline
         print(os.path.basename(uri1))
-        outline(doc, 0, show_called=(doc.workflow is not None))
+        outline(doc, 0, show_called=(doc.workflow is not None), show_all=show_all, shown=shown)
 
     if shellcheck and Lint._shellcheck_available == False:
         print(
-            "* Hint: install shellcheck (www.shellcheck.net) to check task commands. (--no-shellcheck suppresses this message)",
+            "* Suggestion: install shellcheck (www.shellcheck.net) to check task commands. (--no-shellcheck suppresses this message)",
             file=sys.stderr,
         )
 
+    if strict and shown[0]:
+        sys.exit(2)
 
-def outline(obj, level, file=sys.stdout, show_called=True):
+
+def outline(obj, level, file=sys.stdout, show_called=True, show_all=False, shown=None):
     # recursively pretty-print a brief outline of the workflow
     s = "".join(" " for i in range(level * 4))
 
@@ -179,11 +210,14 @@ def outline(obj, level, file=sys.stdout, show_called=True):
     def descend(dobj=None, first_descent=first_descent):
         # show lint for the node just prior to first descent beneath it
         if not first_descent and hasattr(obj, "lint"):
-            for (pos, klass, msg) in sorted(obj.lint, key=lambda t: t[0]):
-                print(
-                    "{}    (Ln {}, Col {}) {}, {}".format(s, pos.line, pos.column, klass, msg),
-                    file=file,
-                )
+            for (pos, klass, msg, suppressed) in sorted(obj.lint, key=lambda t: t[0]):
+                if show_all or not suppressed:
+                    print(
+                        f"{s}    (Ln {pos.line}, Col {pos.column}) {klass}{' (suppressed)' if suppressed else ''}, {msg}",
+                        file=file,
+                    )
+                    if shown:
+                        shown[0] += 1
         first_descent.append(False)
         if dobj:
             outline(
@@ -191,6 +225,8 @@ def outline(obj, level, file=sys.stdout, show_called=True):
                 level + (1 if not isinstance(dobj, Decl) else 0),
                 file=file,
                 show_called=show_called,
+                show_all=show_all,
+                shown=shown,
             )
 
     # document
@@ -353,6 +389,19 @@ def fill_run_subparser(subparsers):
         action="store_true",
         help="upon failure, print error information JSON to standard output (in addition to standard error logging)",
     )
+    group = run_parser.add_argument_group("logging")
+    group.add_argument(
+        "-v",
+        "--verbose",
+        action="store_true",
+        help="increase logging detail & stream tasks' stderr",
+    )
+    group.add_argument(
+        "--no-color",
+        action="store_true",
+        help="disable colored logging and status bar on terminal (also set by NO_COLOR environment variable)",
+    )
+    group.add_argument("--log-json", action="store_true", help="write all logs in JSON")
     group = run_parser.add_argument_group("configuration")
     group.add_argument(
         "--cfg",
@@ -405,17 +454,6 @@ def fill_run_subparser(subparsers):
         action="store_true",
         help="run all containers as the invoking user uid:gid (more secure, but potentially blocks task commands e.g. apt-get)",
     )
-    run_parser.add_argument(
-        "-v",
-        "--verbose",
-        action="store_true",
-        help="increase logging detail & stream tasks' stderr",
-    )
-    run_parser.add_argument(
-        "--no-color",
-        action="store_true",
-        help="disable colored logging on terminal (also set by NO_COLOR environment variable)",
-    )
     # TODO:
     # way to specify None for an optional value (that has a default)
     return run_parser
@@ -440,6 +478,7 @@ def runner(
     as_me=False,
     no_cache=False,
     error_json=False,
+    log_json=False,
     **kwargs,
 ):
     # set up logging
@@ -451,172 +490,194 @@ def runner(
     else:
         logging.raiseExceptions = False
     if kwargs["no_color"]:
-        # picked up by _util.install_coloredlogs()
+        # picked up by _util.configure_logger()
         os.environ["NO_COLOR"] = os.environ.get("NO_COLOR", "")
+    # log_json setting only comes from command line or environment (not cfg file), because we
+    # need it before loading configuration
+    log_json = log_json or (
+        os.environ.get("MINIWDL__LOGGING__JSON", "").lower().strip()
+        in ("t", "y", "1", "true", "yes")
+    )
     logging.basicConfig(level=level)
     logger = logging.getLogger("miniwdl-run")
-    install_coloredlogs(logger)
 
-    # load configuration & apply command-line overrides
-    cfg_arg = None
-    if cfg:
-        assert os.path.isfile(cfg), "--cfg file not found"
-        cfg_arg = [cfg]
-    cfg = runtime.config.Loader(logger, filenames=cfg_arg)
-    cfg_overrides = {
-        "scheduler": {},
-        "file_io": {},
-        "task_runtime": {},
-        "download_cache": {},
-    }
-    if max_tasks is not None:
-        cfg_overrides["scheduler"]["call_concurrency"] = max_tasks
-    if copy_input_files:
-        cfg_overrides["file_io"]["copy_input_files"] = copy_input_files
-    if as_me:
-        cfg_overrides["task_runtime"]["as_user"] = as_me
-    if runtime_defaults:
-        if runtime_defaults.lstrip()[0] == "{":
-            json.loads(runtime_defaults)
-            cfg_overrides["task_runtime"]["defaults"] = runtime_defaults
-        else:
-            with open(runtime_defaults, "r") as infile:
-                cfg_overrides["task_runtime"]["defaults"] = infile.read()
-    if runtime_cpu_max is not None:
-        cfg_overrides["task_runtime"]["cpu_max"] = runtime_cpu_max
-    if runtime_memory_max is not None:
-        runtime_memory_max = (
-            -1 if runtime_memory_max.strip() == "-1" else parse_byte_size(runtime_memory_max)
-        )
-        cfg_overrides["task_runtime"]["memory_max"] = runtime_memory_max
-    if no_cache:
-        cfg_overrides["download_cache"]["get"] = False
+    with ExitStack() as cleanup:
+        set_status = cleanup.enter_context(configure_logger(json=log_json))
 
-    cfg.override(cfg_overrides)
-    cfg.log_all()
-
-    # check root
-    if not path_really_within((run_dir or os.getcwd()), cfg["file_io"]["root"]):
-        logger.error(
-            _(
-                "working directory or --dir must be within the configured `file_io.root' directory",
-                dir=(run_dir or os.getcwd()),
-                root=cfg["file_io"]["root"],
+        if os.geteuid() == 0:
+            logger.warning(
+                "running as root; non-root users should be able to `miniwdl run` as long as they're in the `docker` group"
             )
-        )
-        sys.exit(2)
-    if (
-        cfg["download_cache"].get_bool("get") or cfg["download_cache"].get_bool("put")
-    ) and not path_really_within(cfg["download_cache"]["dir"], cfg["file_io"]["root"]):
-        logger.error(
-            _(
-                "configuration error: 'download_cache.dir' must be within the `file_io.root' directory",
-                dir=cfg["download_cache"]["dir"],
-                root=cfg["file_io"]["root"],
+
+        # load configuration & apply command-line overrides
+        cfg_arg = None
+        if cfg:
+            assert os.path.isfile(cfg), "--cfg file not found"
+            cfg_arg = [cfg]
+        cfg = runtime.config.Loader(logger, filenames=cfg_arg)
+        cfg_overrides = {
+            "scheduler": {},
+            "file_io": {},
+            "task_runtime": {},
+            "download_cache": {},
+            "logging": {"json": log_json},
+        }
+        if max_tasks is not None:
+            cfg_overrides["scheduler"]["call_concurrency"] = max_tasks
+        if copy_input_files:
+            cfg_overrides["file_io"]["copy_input_files"] = copy_input_files
+        if as_me:
+            cfg_overrides["task_runtime"]["as_user"] = as_me
+        if runtime_defaults:
+            if runtime_defaults.lstrip()[0] == "{":
+                json.loads(runtime_defaults)
+                cfg_overrides["task_runtime"]["defaults"] = runtime_defaults
+            else:
+                with open(runtime_defaults, "r") as infile:
+                    cfg_overrides["task_runtime"]["defaults"] = infile.read()
+        if runtime_cpu_max is not None:
+            cfg_overrides["task_runtime"]["cpu_max"] = runtime_cpu_max
+        if runtime_memory_max is not None:
+            runtime_memory_max = (
+                -1 if runtime_memory_max.strip() == "-1" else parse_byte_size(runtime_memory_max)
             )
-        )
-        sys.exit(2)
+            cfg_overrides["task_runtime"]["memory_max"] = runtime_memory_max
+        if no_cache:
+            cfg_overrides["download_cache"]["get"] = False
 
-    # load WDL document
-    doc = load(uri, path or [], check_quant=check_quant, read_source=read_source)
+        cfg.override(cfg_overrides)
+        cfg.log_all()
 
-    # parse and validate the provided inputs
-    def file_found(fn):
-        return runtime.download.able(cfg, fn) or os.path.isfile(fn)
-
-    eff_root = cfg["file_io"]["root"] if not cfg["file_io"].get_bool("copy_input_files") else "/"
-
-    target, input_env, input_json = runner_input(
-        doc,
-        inputs,
-        input_file,
-        empty,
-        task=task,
-        file_found=file_found,
-        root=eff_root,  # if copy_input_files is set, then input files need not reside under the configured root
-    )
-
-    if json_only:
-        print(json.dumps(input_json, indent=2))
-        sys.exit(0)
-
-    # debug logging
-    versionlog = {}
-    for pkg in ["miniwdl", "docker", "lark-parser", "argcomplete", "pygtail"]:
-        try:
-            versionlog[pkg] = str(importlib_metadata.version(pkg))
-        except importlib_metadata.PackageNotFoundError:
-            versionlog[pkg] = "UNKNOWN"
-    logger.debug(_("package versions", **versionlog))
-
-    envlog = {}
-    for k in os.environ:
-        if k.upper().startswith("MINIWDL") or k in [
-            "LANG",
-            "SHELL",
-            "USER",
-            "HOME",
-            "PWD",
-            "TMPDIR",
-        ]:
-            envlog[k] = os.environ[k]
-    logger.debug(_("environment", **envlog))
-
-    enabled_plugins = []
-    disabled_plugins = []
-    for group in runtime.config.DEFAULT_PLUGINS.keys():
-        for enabled, plugin in runtime.config.load_all_plugins(cfg, group):
-            (enabled_plugins if enabled else disabled_plugins).append(
-                f"{plugin.name} = {plugin.value}"
-            )
-    if enabled_plugins or disabled_plugins:
-        logger.debug(_("plugin configuration", enabled=enabled_plugins, disabled=disabled_plugins))
-
-    rerun_sh = f"pushd {shellquote(os.getcwd())} && miniwdl {' '.join(shellquote(t) for t in sys.argv[1:])}; popd"
-
-    # initialize local Docker Swarm
-    runtime.task.SwarmContainer.global_init(cfg, logger)
-
-    # run & log any errors
-    rundir = None
-    try:
-        rundir, output_env = runtime.run(cfg, target, input_env, run_dir=run_dir)
-    except Exception as exn:
-        if error_json:
-            print(json.dumps(runtime.error_json(exn), indent=2))
-        exit_status = 2
-        from_rundir = None
-        while isinstance(exn, runtime.RunFailed):
-            exn_rundir = getattr(exn, "run_dir")
-            rundir = rundir or exn_rundir
-            from_rundir = exn_rundir
-            exn = exn.__cause__
-        if isinstance(exn, runtime.CommandFailed):
-            exit_status = (lambda v: v if v else exit_status)(getattr(exn, "exit_status", 0))
-            if not (kwargs["verbose"] or kwargs["debug"]):
-                logger.notice(
-                    "run with --verbose to include task standard error streams in this log"
+        # check root
+        if not path_really_within((run_dir or os.getcwd()), cfg["file_io"]["root"]):
+            logger.error(
+                _(
+                    "working directory or --dir must be within the configured `file_io.root' directory",
+                    dir=(run_dir or os.getcwd()),
+                    root=cfg["file_io"]["root"],
                 )
-        info = runtime.error_json(exn)
-        if rundir:
-            info["dir"] = rundir
-        if from_rundir and from_rundir != rundir:
-            info["from_dir"] = from_rundir
-        msg = str(exn)
-        if "message" in info:
-            msg = info["message"]
-            del info["message"]
-        logger.error(_(msg, **info))
-        if kwargs["debug"]:
-            raise
-        sys.exit(exit_status)
-    finally:
-        if rundir:
-            # whether success or fail, leave some artifacts
-            with open(os.path.join(rundir, "rerun"), "w") as rerunfile:
-                print(rerun_sh, file=rerunfile)
-            copy_source(doc, os.path.join(rundir, "wdl"))
-        cfg.log_unused_options()
+            )
+            sys.exit(2)
+        if (
+            cfg["download_cache"].get_bool("get") or cfg["download_cache"].get_bool("put")
+        ) and not path_really_within(cfg["download_cache"]["dir"], cfg["file_io"]["root"]):
+            logger.error(
+                _(
+                    "configuration error: 'download_cache.dir' must be within the `file_io.root' directory",
+                    dir=cfg["download_cache"]["dir"],
+                    root=cfg["file_io"]["root"],
+                )
+            )
+            sys.exit(2)
+
+        # load WDL document
+        doc = load(uri, path or [], check_quant=check_quant, read_source=read_source)
+
+        # parse and validate the provided inputs
+        def file_found(fn):
+            return runtime.download.able(cfg, fn) or os.path.isfile(fn)
+
+        eff_root = (
+            cfg["file_io"]["root"] if not cfg["file_io"].get_bool("copy_input_files") else "/"
+        )
+
+        target, input_env, input_json = runner_input(
+            doc,
+            inputs,
+            input_file,
+            empty,
+            task=task,
+            file_found=file_found,
+            root=eff_root,  # if copy_input_files is set, then input files need not reside under the configured root
+        )
+
+        if json_only:
+            print(json.dumps(input_json, indent=2))
+            sys.exit(0)
+
+        # debug logging
+        versionlog = {}
+        for pkg in ["miniwdl", "docker", "lark-parser", "argcomplete", "pygtail"]:
+            try:
+                versionlog[pkg] = str(importlib_metadata.version(pkg))
+            except importlib_metadata.PackageNotFoundError:
+                versionlog[pkg] = "UNKNOWN"
+        logger.debug(_("package versions", **versionlog))
+
+        envlog = {}
+        for k in os.environ:
+            if k.upper().startswith("MINIWDL") or k in [
+                "LANG",
+                "SHELL",
+                "USER",
+                "HOME",
+                "PWD",
+                "TMPDIR",
+            ]:
+                envlog[k] = os.environ[k]
+        logger.debug(_("environment", **envlog))
+
+        enabled_plugins = []
+        disabled_plugins = []
+        for group in runtime.config.DEFAULT_PLUGINS.keys():
+            for enabled, plugin in runtime.config.load_all_plugins(cfg, group):
+                (enabled_plugins if enabled else disabled_plugins).append(
+                    f"{plugin.name} = {plugin.value}"
+                )
+        if enabled_plugins or disabled_plugins:
+            logger.debug(
+                _("plugin configuration", enabled=enabled_plugins, disabled=disabled_plugins)
+            )
+
+        rerun_sh = f"pushd {shellquote(os.getcwd())} && miniwdl {' '.join(shellquote(t) for t in sys.argv[1:])}; popd"
+
+        # initialize local Docker Swarm
+        runtime.task.SwarmContainer.global_init(cfg, logger)
+
+        # run & log any errors
+        cleanup.enter_context(runtime._statusbar.enable(set_status))
+        cache = cleanup.enter_context(runtime.cache.CallCache(cfg, logger))
+        rundir = None
+        try:
+            rundir, output_env = runtime.run(cfg, target, input_env, run_dir=run_dir, _cache=cache)
+        except Exception as exn:
+            if error_json:
+                print(json.dumps(runtime.error_json(exn), indent=2))
+            exit_status = 2
+            from_rundir = None
+            while isinstance(exn, runtime.RunFailed):
+                exn_rundir = getattr(exn, "run_dir")
+                rundir = rundir or exn_rundir
+                from_rundir = exn_rundir
+                exn = exn.__cause__
+            if isinstance(exn, runtime.CommandFailed):
+                exit_status = (lambda v: v if v else exit_status)(getattr(exn, "exit_status", 0))
+                if not (kwargs["verbose"] or kwargs["debug"]):
+                    logger.notice(
+                        "run with --verbose to include task standard error streams in this log"
+                    )
+            info = runtime.error_json(exn)
+            if rundir:
+                info["dir"] = rundir
+            if from_rundir and from_rundir != rundir:
+                info["from_dir"] = from_rundir
+            msg = str(exn)
+            if "message" in info:
+                msg = info["message"]
+                del info["message"]
+            logger.error(_(msg, **info))
+            if kwargs["debug"]:
+                raise
+            sys.exit(exit_status)
+        finally:
+            if rundir:
+                # whether success or fail, leave some artifacts in the run directory.
+                # this should be done under the flock held open within the cache context so that
+                # other waiting processes know when we're really finished with the run directory.
+                with open(os.path.join(rundir, "rerun"), "w") as rerunfile:
+                    print(rerun_sh, file=rerunfile)
+                copy_source(doc, os.path.join(rundir, "wdl"))
+            cfg.log_unused_options()
 
     # report
     outputs_json = {"outputs": values_to_json(output_env, namespace=target.name), "dir": rundir}
@@ -914,6 +975,7 @@ def fill_run_self_test_subparser(subparsers):
         default=None,
         help="configuration file to load (in preference to file named by MINIWDL_CFG environment, or XDG_CONFIG_{HOME,DIRS}/miniwdl.cfg)",
     )
+    run_parser.add_argument("--log-json", action="store_true", help="write all logs in JSON")
     run_parser.add_argument(
         "--as-me", action="store_true", help="run all containers as the current user uid:gid"
     )
@@ -972,7 +1034,7 @@ def run_self_test(**kwargs):
         os.path.join(dn, "test.wdl"),
         "who=https://raw.githubusercontent.com/chanzuckerberg/miniwdl/master/tests/alyssa_ben.txt",
         "--dir",
-        dn,
+        dn if dn not in [".", "./"] else os.getcwd(),
         "--debug",
     ]
     if kwargs["as_me"]:
@@ -980,6 +1042,8 @@ def run_self_test(**kwargs):
     if kwargs["cfg"]:
         argv.append("--cfg")
         argv.append(kwargs["cfg"])
+    if kwargs["log_json"]:
+        argv.append("--log-json")
     try:
         outputs = main(argv)["outputs"]
         assert len(outputs["hello_caller.messages"]) == 2
@@ -1313,120 +1377,120 @@ def localize(
 ):
     logging.basicConfig(level=NOTICE_LEVEL)
     logger = logging.getLogger("miniwdl-localize")
-    install_coloredlogs(logger)
+    with configure_logger() as set_status:
 
-    cfg_arg = None
-    if cfg:
-        assert os.path.isfile(cfg), "--cfg file not found"
-        cfg_arg = [cfg]
-    cfg = runtime.config.Loader(logger, filenames=cfg_arg)
-    cache_cfg = cfg["download_cache"]
-    original_get = cache_cfg.get_bool("get")
-    if original_get and no_cache:
-        cfg.override({"download_cache": {"get": False}})
-    logger.notice(
-        _(
-            "effective configuration",
-            put=cache_cfg.get_bool("put"),
-            get=cache_cfg.get_bool("get"),
-            dir=cache_cfg["dir"],
-            ignore_query=cache_cfg.get_bool("ignore_query"),
-            enable_patterns=cache_cfg.get_list("enable_patterns"),
-            disable_patterns=cache_cfg.get_list("disable_patterns"),
-        )
-    )
-
-    uri = uri or []
-    uri = set(uri)
-
-    if infile:
-        # load WDL document
-        doc = load(wdlfile, path or [], check_quant=check_quant, read_source=read_source)
-
-        def file_found(fn):
-            return runtime.download.able(cfg, fn) or os.path.isfile(fn)
-
-        target, input_env, input_json = runner_input(
-            doc, [], infile, [], task=task, check_required=False, file_found=file_found
-        )
-
-        # scan for Files that appear to be downloadable URIs
-        def scan(x):
-            if isinstance(x, Value.File) and runtime.download.able(cfg, x.value):
-                yield x.value
-            for y in x.children:
-                yield from scan(y)
-
-        for b in input_env:
-            uri |= set(scan(b.value))
-
-    if not uri:
-        logger.warning(
-            "nothing to do; if inputs use special URI schemes, make sure necessary downloader plugin(s) are installed and enabled"
-        )
-        sys.exit(0)
-
-    if not cache_cfg.get_bool("put"):
-        logger.error(
-            'configuration section "download_cache", option "put" (env MINIWDL__DOWNLOAD_CACHE__PUT) must be true for this operation to be effective'
-        )
-        sys.exit(2)
-
-    if not path_really_within(cfg["download_cache"]["dir"], cfg["file_io"]["root"]):
-        logger.error(
+        cfg_arg = None
+        if cfg:
+            assert os.path.isfile(cfg), "--cfg file not found"
+            cfg_arg = [cfg]
+        cfg = runtime.config.Loader(logger, filenames=cfg_arg)
+        cache_cfg = cfg["download_cache"]
+        original_get = cache_cfg.get_bool("get")
+        if original_get and no_cache:
+            cfg.override({"download_cache": {"get": False}})
+        logger.notice(
             _(
-                "configuration error: `download_cache.dir' must be within the `file_io.root' directory",
-                dir=cfg["download_cache"]["dir"],
-                root=cfg["file_io"]["root"],
+                "effective configuration",
+                put=cache_cfg.get_bool("put"),
+                get=cache_cfg.get_bool("get"),
+                dir=cache_cfg["dir"],
+                ignore_query=cache_cfg.get_bool("ignore_query"),
+                enable_patterns=cache_cfg.get_list("enable_patterns"),
+                disable_patterns=cache_cfg.get_list("disable_patterns"),
             )
         )
-        sys.exit(2)
 
-    with runtime.cache.CallCache(cfg, logger) as cache:
-        disabled = [u for u in uri if not cache.download_path(u)]
-    if disabled:
-        logger.notice(_("URIs found but not cacheable per configuration", uri=disabled))
-    uri = list(uri - set(disabled))
+        uri = uri or []
+        uri = set(uri)
 
-    if not uri:
-        logger.warning("nothing to do; check configured enable_patterns and disable_patterns")
-        sys.exit(0)
-    logger.notice(_("starting downloads", uri=uri))
+        if infile:
+            # load WDL document
+            doc = load(wdlfile, path or [], check_quant=check_quant, read_source=read_source)
 
-    # initialize Docker Swarm
-    runtime.task.SwarmContainer.global_init(cfg, logger)
+            def file_found(fn):
+                return runtime.download.able(cfg, fn) or os.path.isfile(fn)
 
-    # cheesy trick: provide the list of URIs as File inputs to a dummy workflow, causing the
-    # runtime to download & cache them
-    localizer_wdl = """
-        version 1.0
-        workflow localize {
-            input {
-                Array[File] uris
+            target, input_env, input_json = runner_input(
+                doc, [], infile, [], task=task, check_required=False, file_found=file_found
+            )
+
+            # scan for Files that appear to be downloadable URIs
+            def scan(x):
+                if isinstance(x, Value.File) and runtime.download.able(cfg, x.value):
+                    yield x.value
+                for y in x.children:
+                    yield from scan(y)
+
+            for b in input_env:
+                uri |= set(scan(b.value))
+
+        if not uri:
+            logger.warning(
+                "nothing to do; if inputs use special URI schemes, make sure necessary downloader plugin(s) are installed and enabled"
+            )
+            sys.exit(0)
+
+        if not cache_cfg.get_bool("put"):
+            logger.error(
+                'configuration section "download_cache", option "put" (env MINIWDL__DOWNLOAD_CACHE__PUT) must be true for this operation to be effective'
+            )
+            sys.exit(2)
+
+        if not path_really_within(cfg["download_cache"]["dir"], cfg["file_io"]["root"]):
+            logger.error(
+                _(
+                    "configuration error: `download_cache.dir' must be within the `file_io.root' directory",
+                    dir=cfg["download_cache"]["dir"],
+                    root=cfg["file_io"]["root"],
+                )
+            )
+            sys.exit(2)
+
+        with runtime.cache.CallCache(cfg, logger) as cache:
+            disabled = [u for u in uri if not cache.download_path(u)]
+        if disabled:
+            logger.notice(_("URIs found but not cacheable per configuration", uri=disabled))
+        uri = list(uri - set(disabled))
+
+        if not uri:
+            logger.warning("nothing to do; check configured enable_patterns and disable_patterns")
+            sys.exit(0)
+        logger.notice(_("starting downloads", uri=uri))
+
+        # initialize Docker Swarm
+        runtime.task.SwarmContainer.global_init(cfg, logger)
+
+        # cheesy trick: provide the list of URIs as File inputs to a dummy workflow, causing the
+        # runtime to download & cache them
+        localizer_wdl = """
+            version 1.0
+            workflow localize {
+                input {
+                    Array[File] uris
+                }
+                output {
+                    Array[File] files = uris
+                }
             }
-            output {
-                Array[File] files = uris
-            }
-        }
-        """
-    localizer = parse_document(localizer_wdl)
-    localizer.typecheck()
-    cfg = runtime.config.Loader(logger)
-    subdir, outputs = runtime.run(
-        cfg,
-        localizer.workflow,
-        values_from_json({"uris": uri}, localizer.workflow.available_inputs),
-        run_dir=os.environ.get("TMPDIR", "/tmp"),
-    )
-
-    logger.notice(
-        _("success", files=[os.path.realpath(p) for p in values_to_json(outputs)["files"]])
-    )
-    if not original_get:
-        logger.warning(
-            """future runs won't use the cache unless configuration section "download_cache", key "get" (env MINIWDL__DOWNLOAD_CACHE__GET) is set to true"""
+            """
+        localizer = parse_document(localizer_wdl)
+        localizer.typecheck()
+        cfg = runtime.config.Loader(logger)
+        subdir, outputs = runtime.run(
+            cfg,
+            localizer.workflow,
+            values_from_json({"uris": uri}, localizer.workflow.available_inputs),
+            run_dir=os.environ.get("TMPDIR", "/tmp"),
         )
 
+        logger.notice(
+            _("success", files=[os.path.realpath(p) for p in values_to_json(outputs)["files"]])
+        )
+        if not original_get:
+            logger.warning(
+                """future runs won't use the cache unless configuration section "download_cache", key "get" (env MINIWDL__DOWNLOAD_CACHE__GET) is set to true"""
+            )
+
 
 def die(msg, status=2):
     print("\n" + msg + "\n", file=sys.stderr)
