@@ -10,6 +10,8 @@ import threading
 import time
 import copy
 import fcntl
+import subprocess
+import shutil
 from time import sleep
 from datetime import datetime
 from contextlib import contextmanager, AbstractContextManager
@@ -30,6 +32,7 @@ from typing import (
 )
 from types import FrameType
 import coloredlogs
+from pythonjsonlogger import jsonlogger
 from pygtail import Pygtail
 import docker
 
@@ -183,36 +186,43 @@ def write_values_json(
 
 
 @export
-def provision_run_dir(name: str, run_dir: Optional[str] = None) -> str:
+def provision_run_dir(name: str, parent_dir: Optional[str], last_link: bool = False) -> str:
     here = (
-        (run_dir in [".", "./"] or run_dir.endswith("/.") or run_dir.endswith("/./"))
-        if run_dir
+        (parent_dir in [".", "./"] or parent_dir.endswith("/.") or parent_dir.endswith("/./"))
+        if parent_dir
         else False
     )
-    run_dir = os.path.abspath(run_dir or os.getcwd())
+    parent_dir = os.path.abspath(parent_dir or os.getcwd())
 
+    run_dir = None
     if here:
+        # user wants to use parent_dir exactly
+        run_dir = parent_dir
         os.makedirs(run_dir, exist_ok=True)
-        return run_dir
+        parent_dir = os.path.dirname(parent_dir)
+    else:
+        # create timestamp-named directory
+        while not run_dir:
+            run_dir = os.path.join(
+                parent_dir, datetime.today().strftime("%Y%m%d_%H%M%S") + "_" + name
+            )
+            try:
+                os.makedirs(run_dir, exist_ok=False)
+            except FileExistsError:
+                run_dir = None
+                sleep(0.5)
+    assert run_dir
 
-    now = datetime.today()
-    run_dir2 = os.path.join(run_dir, now.strftime("%Y%m%d_%H%M%S") + "_" + name)
-    try:
-        os.makedirs(run_dir2, exist_ok=False)
-        return run_dir2
-    except FileExistsError:
-        pass
+    # update the _LAST link
+    if last_link and run_dir != parent_dir:
+        last_link_name = os.path.join(parent_dir, "_LAST")
+        if os.path.islink(last_link_name) or not (here or os.path.lexists(last_link_name)):
+            run_dir_basename = os.path.basename(run_dir)
+            tmp_link_name = last_link_name + "." + run_dir_basename
+            os.symlink(run_dir_basename, tmp_link_name)
+            os.rename(tmp_link_name, last_link_name)
 
-    while True:
-        run_dir2 = os.path.join(
-            run_dir,
-            now.strftime("%Y%m%d_%H%M%S_") + str(int(now.microsecond / 1000)).zfill(3) + "_" + name,
-        )
-        try:
-            os.makedirs(run_dir2, exist_ok=False)
-            return run_dir2
-        except FileExistsError:
-            sleep(1e-3)
+    return run_dir
 
 
 @export
@@ -231,6 +241,28 @@ class StructuredLogMessage:
         )
 
 
+class StructuredLogMessageJSONFormatter(jsonlogger.JsonFormatter):
+    "JSON formatter for StructuredLogMessages"
+
+    def format(self, rec: logging.LogRecord) -> str:
+        if isinstance(rec.msg, StructuredLogMessage):
+            ans = {"level": rec.levelname, "message": rec.msg.message}
+            for k, v in rec.msg.kwargs.items():
+                if k not in ans:
+                    ans[k] = v
+            rec.msg = ans
+        return super().format(rec)
+
+    def add_fields(
+        self, log_record: Dict[str, Any], record: logging.LogRecord, message_dict: Dict[str, Any]
+    ) -> None:
+        super().add_fields(log_record, record, message_dict)
+        log_record["timestamp"] = round(record.created, 3)
+        log_record["source"] = record.name
+        log_record["level"] = record.levelname
+        log_record["levelno"] = record.levelno
+
+
 VERBOSE_LEVEL = 15
 __all__.append("VERBOSE_LEVEL")
 logging.addLevelName(VERBOSE_LEVEL, "VERBOSE")
@@ -254,63 +286,168 @@ def notice(self, message, *args, **kws):  # pyre-fixme
 
 logging.Logger.notice = notice
 
+
+@export
+class ANSI:
+    # https://gist.github.com/RabaDabaDoba/145049536f815903c79944599c6f952a
+    # https://espterm.github.io/docs/VT100%20escape%20codes.html
+    CLEAR: str = "\x1b[2K\r"
+    RESET: str = "\x1b[0m"
+    BOLD: str = "\x1b[1m"
+
+    RED: str = "\x1b[0;31m"
+    BRED: str = "\x1b[1;31m"
+    HRED: str = "\x1b[0;91m"
+    BHRED: str = "\x1b[1;91m"
+
+    HIDE_CURSOR: str = "\x1b[?25l"
+    SHOW_CURSOR: str = "\x1b[?25h"
+
+
+def _ansilen(parts: List[str]) -> int:
+    return sum([len(s) for s in parts if s[0] != "\x1b"])
+
+
+class _StatusLineStandardErrorHandler(coloredlogs.StandardErrorHandler):
+    """
+    This subclass augments coloredlogs.StandardErrorHandler to maintain a "status line" which
+    remains in place at the bottom of the screen as log records scroll by. The content of the
+    status line can be set at any time. It will be truncated to the terminal width.
+    """
+
+    _singleton: "Optional[_StatusLineStandardErrorHandler]" = None
+    _status: str = ""
+
+    def __init__(self, *args, **kwargs):  # pyre-ignore
+        super().__init__(*args, **kwargs)
+        assert not self.__class__._singleton
+        self.__class__._singleton = self
+
+    def emit(self, record: logging.LogRecord) -> None:
+        self.acquire()
+        try:
+            sys.stderr.write(ANSI.CLEAR)
+            super().emit(record)
+            self.emit_status()
+        finally:
+            self.release()
+
+    def emit_status(self) -> None:
+        self.acquire()
+        try:
+            sys.stderr.write(ANSI.CLEAR + self._status + ANSI.RESET)
+            self.flush()
+        finally:
+            self.release()
+
+    def set_status(self, new_status: List[str]) -> None:
+        cols = shutil.get_terminal_size().columns
+        if _ansilen(new_status) > cols:
+            new_status = new_status.copy()
+            while new_status and (_ansilen(new_status) > cols or new_status[-1][0] == "\x1b"):
+                new_status.pop()
+        self._status = "".join(new_status)
+        self.emit_status()
+
+
 LOGGING_FORMAT = "%(asctime)s.%(msecs)03d %(name)s %(levelname)s %(message)s"
+COLORED_LOGGING_FORMAT = "%(asctime)s.%(msecs)03d %(name)s %(message)s"  # colors obviate levelname
 __all__.append("LOGGING_FORMAT")
 
 
 @export
-def install_coloredlogs(logger: logging.Logger) -> None:
-    level_styles = {}
-    field_styles = {}
+@contextmanager
+def configure_logger(
+    force_tty: bool = False, json: bool = False
+) -> Iterator[Callable[[str], None]]:
+    """
+    contextmanager to set up the root/stderr logger; yields a function to set the status line at
+    the bottom of the screen (if stderr isatty, else it does nothing)
+    """
+    logger = logging.getLogger()
 
-    if sys.stderr.isatty() and "NO_COLOR" not in os.environ:
-        level_styles = dict(coloredlogs.DEFAULT_LEVEL_STYLES)
-        level_styles["debug"]["color"] = 242
-        level_styles["notice"] = {"color": "green", "bold": True}
-        level_styles["error"]["bold"] = True
-        level_styles["warning"]["bold"] = True
-        level_styles["info"] = {}
-        field_styles = dict(coloredlogs.DEFAULT_FIELD_STYLES)
-        field_styles["asctime"] = {"color": "blue"}
-        field_styles["name"] = {"color": "magenta"}
+    if json:
+        logger.handlers[0].setFormatter(StructuredLogMessageJSONFormatter())
+        yield (lambda ignore: None)
+    else:
+        level_styles = {}
+        field_styles = {}
+        fmt = LOGGING_FORMAT
+        tty = force_tty or (sys.stderr.isatty() and "NO_COLOR" not in os.environ)
 
-    coloredlogs.install(
-        level=logger.getEffectiveLevel(),
-        logger=logger,
-        level_styles=level_styles,
-        field_styles=field_styles,
-        fmt=LOGGING_FORMAT,
-    )
+        if tty:
+            level_styles = dict(coloredlogs.DEFAULT_LEVEL_STYLES)
+            level_styles["debug"]["color"] = 242
+            level_styles["notice"] = {"color": "green", "bold": True}
+            level_styles["error"]["bold"] = True
+            level_styles["warning"]["bold"] = True
+            level_styles["info"] = {}
+            field_styles = dict(coloredlogs.DEFAULT_FIELD_STYLES)
+            field_styles["asctime"] = {"color": "blue"}
+            field_styles["name"] = {"color": "magenta"}
+            fmt = COLORED_LOGGING_FORMAT
+
+            # monkey-patch _StatusLineStandardErrorHandler over coloredlogs.StandardErrorHandler for
+            # coloredlogs.install() to instantiate
+            coloredlogs.StandardErrorHandler = _StatusLineStandardErrorHandler
+            sys.stderr.write(ANSI.HIDE_CURSOR)  # hide cursor
+
+        try:
+            coloredlogs.install(
+                level=logger.getEffectiveLevel(),
+                logger=logger,
+                level_styles=level_styles,
+                field_styles=field_styles,
+                fmt=fmt,
+            )
+            yield (
+                lambda status: _StatusLineStandardErrorHandler._singleton.set_status(  # pyre-fixme
+                    status
+                )
+                if _StatusLineStandardErrorHandler._singleton
+                else None
+            )
+        finally:
+            if tty:
+                sys.stderr.write(ANSI.CLEAR)  # wipe the status line
+                sys.stderr.write(ANSI.SHOW_CURSOR)  # un-hide cursor
 
 
 @export
 @contextmanager
 def PygtailLogger(
-    logger: logging.Logger, filename: str, prefix: str = "2| "
+    logger: logging.Logger, filename: str, callback: Optional[Callable[[str], None]] = None
 ) -> Iterator[Callable[[], None]]:
     """
     Helper for streaming task stderr into logger using pygtail. Context manager yielding a function
     which reads the latest lines from the file and writes them into logger at verbose level. This
     function also runs automatically on context exit.
 
-    Truncates lines at 4KB in case writer goes haywire.
+    Stops if it sees a line greater than 4KB, in case writer goes haywire.
     """
     pygtail = None
-    if logger.getEffectiveLevel() <= VERBOSE_LEVEL:
+    if logger.isEnabledFor(VERBOSE_LEVEL):
         pygtail = Pygtail(filename, full_lines=True)
+    logger2 = logger.getChild("stderr")
+
+    def default_callback(line: str) -> None:
+        assert len(line) <= 4096, "line > 4KB"
+        logger2.verbose(line.rstrip())  # pyre-fixme
+
+    callback = callback or default_callback
 
     def poll() -> None:
         nonlocal pygtail
         if pygtail:
             try:
                 for line in pygtail:
-                    logger.verbose((prefix + line.rstrip())[:4096])  # pyre-ignore
-            except:
+                    callback(line)
+            except Exception as exn:
                 # cf. https://github.com/bgreenlee/pygtail/issues/48
-                logger.verbose(  # pyre-ignore
-                    "incomplete log stream due to the following exception; see %s",
-                    filename,
-                    exc_info=sys.exc_info(),
+                logger.warning(
+                    StructuredLogMessage(
+                        "log stream is incomplete", filename=filename, error=str(exn)
+                    )
                 )
                 pygtail = None
 
@@ -499,12 +636,19 @@ def chmod_R_plus(path: str, file_bits: int = 0, dir_bits: int = 0) -> None:
 
 @export
 @contextmanager
-def LoggingFileHandler(logger: logging.Logger, filename: str) -> Iterator[logging.FileHandler]:
+def LoggingFileHandler(
+    logger: logging.Logger, filename: str, json: bool = False
+) -> Iterator[logging.FileHandler]:
     """
     Context manager which opens a logging.FileHandler and adds it to the logger; on exit, closes
     the log file and removes the handler.
     """
     fh = logging.FileHandler(filename)
+    fh.setFormatter(
+        StructuredLogMessageJSONFormatter()
+        if json
+        else logging.Formatter(LOGGING_FORMAT, datefmt="%Y-%m-%d %H:%M:%S")
+    )
     try:
         logger.addHandler(fh)
         yield fh
@@ -514,21 +658,6 @@ def LoggingFileHandler(logger: logging.Logger, filename: str) -> Iterator[loggin
         logger.removeHandler(fh)
 
 
-@export
-class AtomicCounter:
-    _value: int
-    _lock: threading.Lock
-
-    def __init__(self) -> None:
-        self._value = 0
-        self._lock = threading.Lock()
-
-    def next(self) -> int:
-        with self._lock:
-            self._value += 1
-            return self._value
-
-
 @export
 @contextmanager
 def compose_coroutines(  # pyre-fixme
@@ -700,8 +829,9 @@ class FlockHolder(AbstractContextManager):
                             openfile.fileno(), ns=(int(time.time() * 1e9), file_st.st_mtime_ns)
                         )
 
-                    # The filename link could have been replaced or removed in the instant between
-                    # our open() and flock() syscalls.
+                    # Even if all concurrent processes obey the advisory flocks, the filename link
+                    # could have been replaced or removed in the duration between our open() and
+                    # fcntl() syscalls.
                     # - if it was removed, the following os.stat will trigger FileNotFoundError,
                     #   which is reasonable to propagate.
                     # - if it was replaced, the subsequent condition won't hold, and we'll loop
@@ -728,3 +858,10 @@ class FlockHolder(AbstractContextManager):
                     openfile.close()
                     raise
                 openfile.close()
+
+
+@export
+class RepeatTimer(threading.Timer):
+    def run(self) -> None:
+        while not self.finished.wait(self.interval):  # pyre-ignore
+            self.function(*self.args, **self.kwargs)  # pyre-ignore
