@@ -16,7 +16,10 @@ import threading
 import shutil
 import shlex
 import re
-import socket
+import hashlib
+import uuid
+import base64
+import contextlib
 from abc import ABC, abstractmethod
 from typing import Tuple, List, Dict, Optional, Callable, Iterable, Set, Any
 from contextlib import ExitStack
@@ -26,18 +29,16 @@ from .._util import (
     write_atomic,
     write_values_json,
     provision_run_dir,
-    LOGGING_FORMAT,
     PygtailLogger,
     TerminationSignalFlag,
     parse_byte_size,
     chmod_R_plus,
     path_really_within,
     LoggingFileHandler,
-    AtomicCounter,
     compose_coroutines,
 )
 from .._util import StructuredLogMessage as _
-from . import config
+from . import config, _statusbar
 from .download import able as downloadable, run_cached as download
 from .cache import CallCache
 from .error import *
@@ -99,6 +100,13 @@ class TaskContainer(ABC):
 
     input_file_map_rev: Dict[str, str]
 
+    stderr_callback: Optional[Callable[[str], None]]
+    """
+    A function called line-by-line for the task's standard error stream, iff verbose logging is
+    enabled. If provided by a plugin then it overrides the default standard error logging, which
+    writes each line to the 'stderr' child of the task logger.
+    """
+
     _running: bool
 
     def __init__(self, cfg: config.Loader, run_id: str, host_dir: str) -> None:
@@ -108,6 +116,7 @@ class TaskContainer(ABC):
         self.container_dir = "/mnt/miniwdl_task_container"
         self.input_file_map = {}
         self.input_file_map_rev = {}
+        self.stderr_callback = None
         self._running = False
         os.makedirs(os.path.join(self.host_dir, "work"))
 
@@ -162,16 +171,25 @@ class TaskContainer(ABC):
             os.makedirs(os.path.dirname(host_copy_filename), exist_ok=True)
             shutil.copy(host_filename, host_copy_filename)
 
-    def run(self, logger: logging.Logger, command: str, cpu: int, memory: int) -> None:
+    def run(
+        self,
+        logger: logging.Logger,
+        command: str,
+        cpu: int,
+        memory_reservation: int,
+        memory_limit: int,
+    ) -> None:
         """
-        1. Container is instantiated with the configured mounts
+        1. Container is instantiated with the configured mounts and resources
         2. The mounted directory and all subdirectories have u+rwx,g+rwx permission bits; all files
            within have u+rw,g+rw permission bits.
         3. Command is executed in ``{host_dir}/work/`` (where {host_dir} is mounted to
            {container_dir} inside the container)
         4. Standard output is written to ``{host_dir}/stdout.txt``
         5. Standard error is written to ``{host_dir}/stderr.txt`` and logged at VERBOSE level
-        6. Raises CommandFailed for nonzero exit code, or any other error
+        6. Raises CommandFailed for nonzero exit code
+        7. Raises Terminated if TerminationSignalFlag detected, or Interrupted if the backend
+           cancels on us for some reason that isn't our fault.
 
         The container is torn down in any case, including SIGTERM/SIGHUP signal which is trapped.
         """
@@ -181,10 +199,12 @@ class TaskContainer(ABC):
         if command.strip():  # if the command is empty then don't bother with any of this
             with TerminationSignalFlag(logger) as terminating:
                 if terminating():
-                    raise Terminated()
+                    raise Terminated(quiet=True)
                 self._running = True
                 try:
-                    exit_status = self._run(logger, terminating, command, cpu, memory)
+                    exit_status = self._run(
+                        logger, terminating, command, cpu, memory_reservation, memory_limit
+                    )
                 finally:
                     self._running = False
 
@@ -200,25 +220,31 @@ class TaskContainer(ABC):
         terminating: Callable[[], bool],
         command: str,
         cpu: int,
-        memory: int,
+        memory_reservation: int,
+        memory_limit: int,
     ) -> int:
         # run command in container & return exit status
         raise NotImplementedError()
 
-    def reset(self, logger: logging.Logger, prev_retries: int) -> None:
+    def reset(self, logger: logging.Logger, retries: int, delete_work: bool = False) -> None:
         """
         After a container/command failure, reset the working directory state so that
         copy_input_files() and run() can be retried.
         """
-        artifacts_dir = os.path.join(self.host_dir, "failed_tries", str(prev_retries))
-        artifacts_moved = []
+        artifacts_dir = os.path.join(self.host_dir, "failed_tries", str(retries))
+        artifacts = []
         for artifact in ["work", "command", "stdout.txt", "stderr.txt", "stderr.txt.offset"]:
             src = os.path.join(self.host_dir, artifact)
             if os.path.exists(src):
-                os.renames(src, os.path.join(artifacts_dir, artifact))
-                artifacts_moved.append(src)
+                artifacts.append(src)
+                if delete_work:
+                    (shutil.rmtree if os.path.isdir(src) else os.unlink)(src)
+                else:
+                    os.renames(src, os.path.join(artifacts_dir, artifact))
         logger.info(
-            _("archived failed task artifacts", artifacts=artifacts_moved, dest=artifacts_dir)
+            _("deleted failed task artifacts", artifacts=artifacts)
+            if delete_work
+            else _("archived failed task artifacts", artifacts=artifacts, dest=artifacts_dir)
         )
         os.makedirs(os.path.join(self.host_dir, "work"))
 
@@ -269,7 +295,6 @@ class SwarmContainer(TaskContainer):
     """
 
     _limits: Dict[str, int] = {}
-    _id_counter: AtomicCounter = AtomicCounter()
 
     @classmethod
     def global_init(cls, cfg: config.Loader, logger: logging.Logger) -> None:
@@ -305,7 +330,12 @@ class SwarmContainer(TaskContainer):
                     logger.warning(
                         "docker swarm is inactive on this host; performing `docker swarm init --advertise-addr 127.0.0.1 --listen-addr 127.0.0.1`"
                     )
-                    client.swarm.init(advertise_addr="127.0.0.1", listen_addr="127.0.0.1")
+                    try:
+                        client.swarm.init(advertise_addr="127.0.0.1", listen_addr="127.0.0.1")
+                    except Exception as exn:
+                        # smooth over race condition with multiple processes trying to init swarm
+                        if "already part of a swarm" not in str(exn):
+                            raise exn
 
                 logger.notice(  # pyre-fixme
                     _(
@@ -403,7 +433,8 @@ class SwarmContainer(TaskContainer):
         terminating: Callable[[], bool],
         command: str,
         cpu: int,
-        memory: int,
+        memory_reservation: int,
+        memory_limit: int,
     ) -> int:
         self._observed_states = set()
         with open(os.path.join(self.host_dir, "command"), "x") as outfile:
@@ -423,7 +454,9 @@ class SwarmContainer(TaskContainer):
 
         # connect to dockerd
         client = docker.from_env(timeout=900)
-        resources, user, groups = self.misc_config(logger, client, cpu, memory)
+        resources, user, groups = self.misc_config(
+            logger, client, cpu, memory_reservation, memory_limit
+        )
         svc = None
         exit_code = None
         try:
@@ -431,7 +464,7 @@ class SwarmContainer(TaskContainer):
             # scheduling (waiting until requested # of CPUs are available).
             kwargs = {
                 # unique name with some human readability; docker limits to 63 chars (issue #327)
-                "name": f"wdl-{os.getpid()}-{SwarmContainer._id_counter.next()}-{self.run_id}"[:63],
+                "name": self.unique_service_name(self.run_id),
                 "command": [
                     "/bin/bash",
                     "-c",
@@ -453,18 +486,34 @@ class SwarmContainer(TaskContainer):
             logger.debug(_("docker service", name=svc.name, id=svc.short_id))
 
             # stream stderr into log
-            with PygtailLogger(logger, os.path.join(self.host_dir, "stderr.txt")) as poll_stderr:
+            with contextlib.ExitStack() as cleanup:
+                poll_stderr = cleanup.enter_context(
+                    PygtailLogger(
+                        logger,
+                        os.path.join(self.host_dir, "stderr.txt"),
+                        callback=self.stderr_callback,
+                    )
+                )
+
                 # poll for container exit
                 while exit_code is None:
-                    # spread out work over the GIL
-                    # TODO: adaptive interval before container starts running (poll less frequently
-                    # if it's already been waiting a long time)
-                    time.sleep(random.uniform(1.0, 2.0))
+                    time.sleep(random.uniform(1.0, 2.0))  # spread out work over the GIL
                     if terminating():
-                        raise Terminated() from None
-                    if "running" in self._observed_states:
-                        poll_stderr()
+                        quiet = not self._observed_states.difference(
+                            # reduce log noise if the terminated task only sat in docker's queue
+                            {"(UNKNOWN)", "new", "allocated", "pending"}
+                        )
+                        if not quiet:
+                            self.poll_service(logger, svc, verbose=True)
+                        raise Terminated(quiet=quiet)
+                    was_running = "running" in self._observed_states
                     exit_code = self.poll_service(logger, svc)
+                    if "running" in self._observed_states:
+                        if not was_running:
+                            # indicate actual container start in status bar
+                            cleanup.enter_context(_statusbar.task_running(cpu, memory_reservation))
+                        poll_stderr()
+
                 logger.debug(
                     _(
                         "docker service logs",
@@ -534,15 +583,22 @@ class SwarmContainer(TaskContainer):
         return mounts
 
     def misc_config(
-        self, logger: logging.Logger, client: docker.DockerClient, cpu: int, memory: int
+        self,
+        logger: logging.Logger,
+        client: docker.DockerClient,
+        cpu: int,
+        memory_reservation: int,
+        memory_limit: int,
     ) -> Tuple[Optional[Dict[str, str]], Optional[str], List[str]]:
         resources = {}
         if cpu > 0:
             # the cpu unit expected by swarm is "NanoCPUs"
             resources["cpu_limit"] = cpu * 1_000_000_000
             resources["cpu_reservation"] = cpu * 1_000_000_000
-        if memory > 0:
-            resources["mem_reservation"] = memory
+        if memory_reservation > 0:
+            resources["mem_reservation"] = memory_reservation
+        if memory_limit > 0:
+            resources["mem_limit"] = memory_limit
         if resources:
             logger.debug(_("docker resources", **resources))
             resources = docker.types.Resources(**resources)
@@ -566,49 +622,78 @@ class SwarmContainer(TaskContainer):
         return resources, user, groups
 
     def poll_service(
-        self, logger: logging.Logger, svc: docker.models.services.Service
+        self, logger: logging.Logger, svc: docker.models.services.Service, verbose: bool = False
     ) -> Optional[int]:
         status = {"State": "(UNKNOWN)"}
 
         svc.reload()
         assert svc.attrs["Spec"]["Labels"]["miniwdl_run_id"] == self.run_id
         tasks = svc.tasks()
+        # logger.debug(_("xxx", service=svc.attrs, tasks=tasks))
         if tasks:
             assert len(tasks) == 1, "docker service should have at most 1 task"
             status = tasks[0]["Status"]
-            logger.debug(_("docker task status", **status))
+            status["DesiredState"] = tasks[0].get("DesiredState", None)
+            if verbose:
+                logger.info(_("docker task status", **status))
+            elif logger.isEnabledFor(logging.DEBUG):
+                logger.debug(_("docker task status", **status))
         else:
             assert (
                 len(self._observed_states or []) <= 1
             ), "docker task shouldn't disappear from service"
 
+        # references on docker task states:
+        # https://docs.docker.com/engine/swarm/how-swarm-mode-works/swarm-task-states/
+        # https://github.com/docker/swarmkit/blob/master/design/task_model.md
+        # https://github.com/moby/moby/blob/8fbf2598f58fb212230e6ddbcfbde628b0458250/api/types/swarm/task.go#L12
+
         # log each new state
-        assert isinstance(self._observed_states, set)
-        if status["State"] not in self._observed_states:
+        state = status["State"]
+        assert isinstance(state, str) and isinstance(self._observed_states, set)
+        if state not in self._observed_states:
             loginfo = {"service": svc.short_id}
             if tasks:
                 loginfo["task"] = tasks[0]["ID"][:10]
                 if "NodeID" in tasks[0]:
                     loginfo["node"] = tasks[0]["NodeID"][:10]
-            method = logger.notice if status["State"] == "running" else logger.info  # pyre-fixme
-            method(_(f"docker task {status['State']}", **loginfo))
-            self._observed_states.add(status["State"])
+            if status["DesiredState"] != state:
+                loginfo["desired"] = status["DesiredState"]
+            logmsg = status.get("Err", status.get("Message", None))
+            if logmsg and logmsg != state:
+                loginfo["message"] = logmsg
+            method = logger.notice if state == "running" else logger.info  # pyre-fixme
+            method(_(f"docker task {state}", **loginfo))
+            self._observed_states.add(state)
 
-        # https://docs.docker.com/engine/swarm/how-swarm-mode-works/swarm-task-states/
-        # https://github.com/moby/moby/blob/8fbf2598f58fb212230e6ddbcfbde628b0458250/api/types/swarm/task.go#L12
+        # determine whether docker task has exited
+        exit_code = None
         if "ExitCode" in status.get("ContainerStatus", {}):
-            exit_code = status["ContainerStatus"]["ExitCode"]
+            exit_code = status["ContainerStatus"]["ExitCode"]  # pyre-fixme
             assert isinstance(exit_code, int)
-            if exit_code != 0 or status["State"] == "complete":
-                logger.notice(  # pyre-fixme
-                    _("docker task exit", state=status["State"], exit_code=exit_code)
-                )
-                return exit_code
 
-        if status["State"] in ["failed", "rejected", "orphaned", "remove"]:
-            raise RuntimeError(
-                f"docker task {status['State']}"
-                + ((": " + status["Err"]) if "Err" in status else "")
+        if state in ("complete", "failed"):
+            logger.notice(_("docker task exit", state=state, exit_code=exit_code))  # pyre-fixme
+            assert isinstance(exit_code, int) and (exit_code == 0) == (state == "complete")
+            return exit_code
+        elif {state, status["DesiredState"]}.intersection(
+            {"rejected", "shutdown", "orphaned", "remove"}
+        ) or exit_code not in [None, 0]:
+            # "rejected" state usually arises from nonexistent docker image.
+            # if the worker assigned a task goes down, any of the following can manifest:
+            #   - exit_code=-1 with state running (or other non-terminal)
+            #   - state shutdown, orphaned, remove
+            #   - desired_state shutdown
+            # also see GitHub issue #374
+            raise (RuntimeError if state == "rejected" else Interrupted)(  # pyre-ignore
+                f"docker task {state}"
+                + (
+                    (", desired state " + status["DesiredState"])
+                    if status["DesiredState"] not in (None, state)
+                    else ""
+                )
+                + (f", exit code = {exit_code}" if exit_code not in [None, 0] else "")
+                + (f": {status['Err']}" if "Err" in status else "")
             )
 
         return None
@@ -631,9 +716,7 @@ class SwarmContainer(TaskContainer):
                 try:
                     chowner = client.containers.run(
                         "alpine:3",
-                        name=f"wdl-chown-{os.getpid()}-{SwarmContainer._id_counter.next()}-{self.run_id}"[
-                            :63
-                        ],
+                        name=self.unique_service_name("chown-" + self.run_id),
                         command=["/bin/ash", "-c", script],
                         volumes=volumes,
                         detach=True,
@@ -651,6 +734,24 @@ class SwarmContainer(TaskContainer):
                     raise
                 logger.exception("post-task chown also failed")
 
+    def unique_service_name(self, run_id: str) -> str:
+        # We need to give each service a name unique on the swarm; collisions cause the service
+        # create request to fail. Considerations:
+        # 1. [0-9A-Za-z-]{1,63} -- case is remembered, but comparison ignores it.
+        # 2. It's useful for the names to be mostly human-readable via `docker service ls` to get a
+        #    sense of what's happening on the swarm. Unfortunately, that tool truncates the display
+        #    names pretty short, so prefer human-readability of the leftmost part of the name.
+        # 3. PID+seqno isn't sufficient because the swarm could receive submissions from miniwdl
+        #    running in different hosts/VMs/containers with potentially colliding PIDs.
+        # see GitHub issues: 327, 368
+        junk = hashlib.sha256()
+        junk.update(uuid.uuid1().bytes)
+        junk.update(uuid.uuid4().bytes)
+        junk = junk.digest()[:15]
+        junk = base64.b32encode(junk).decode().lower()
+        assert len(junk) == 24
+        return f"wdl-{run_id[:34]}-{junk}"  # 4 + 34 + 1 + 24 = 63
+
 
 def run_local_task(
     cfg: config.Loader,
@@ -675,17 +776,28 @@ def run_local_task(
                     If the final path component is ".", then operate in run_dir directly.
     """
 
-    # provision run directory and log file
-    run_id = run_id or task.name
     _run_id_stack = _run_id_stack or []
-    run_dir = provision_run_dir(task.name, run_dir)
-
+    run_id = run_id or task.name
     logger_prefix = (logger_prefix or ["wdl"]) + ["t:" + run_id]
     logger = logging.getLogger(".".join(logger_prefix))
-    logfile = os.path.join(run_dir, "task.log")
     with ExitStack() as cleanup:
-        fh = cleanup.enter_context(LoggingFileHandler(logger, logfile))  # pylint: disable=no-member
-        fh.setFormatter(logging.Formatter(LOGGING_FORMAT))
+        cleanup.enter_context(_statusbar.task_slotted())
+        terminating = cleanup.enter_context(TerminationSignalFlag(logger))
+        if terminating():
+            raise Terminated(quiet=True)
+
+        # provision run directory and log file
+        run_dir = provision_run_dir(task.name, run_dir, last_link=not _run_id_stack)
+        logfile = os.path.join(run_dir, "task.log")
+        fh = cleanup.enter_context(
+            LoggingFileHandler(
+                logger,
+                logfile,
+                json=(
+                    cfg["logging"].get_bool("json") if cfg.has_option("logging", "json") else False
+                ),
+            )
+        )
         logger.notice(  # pyre-fixme
             _(
                 "task start",
@@ -700,8 +812,7 @@ def run_local_task(
         write_values_json(inputs, os.path.join(run_dir, "inputs.json"))
 
         if not _run_id_stack:
-            assert not _cache
-            cache = cleanup.enter_context(CallCache(cfg, logger))  # pylint: disable=no-member
+            cache = _cache or cleanup.enter_context(CallCache(cfg, logger))
             cache.flock(logfile, exclusive=True)  # no containing workflow; flock task.log
         else:
             cache = _cache
@@ -755,7 +866,7 @@ def run_local_task(
                 command, runtime, container = (recv[k] for k in ("command", "runtime", "container"))
 
                 # start container & run command (and retry if needed)
-                _try_task(cfg, logger, container, command, runtime)
+                _try_task(cfg, logger, container, command, runtime, terminating)
 
                 # evaluate output declarations
                 outputs = _eval_task_outputs(logger, task, container_env, container)
@@ -783,7 +894,11 @@ def run_local_task(
         except Exception as exn:
             logger.debug(traceback.format_exc())
             wrapper = RunFailed(task, run_id, run_dir)
-            logger.error(_(str(wrapper), dir=run_dir, **error_json(exn)))
+            logmsg = _(str(wrapper), dir=run_dir, **error_json(exn))
+            if isinstance(exn, Terminated) and getattr(exn, "quiet", False):
+                logger.debug(logmsg)
+            else:
+                logger.error(logmsg)
             try:
                 write_atomic(
                     json.dumps(error_json(wrapper, cause=exn), indent=2),
@@ -795,7 +910,7 @@ def run_local_task(
             try:
                 _delete_work(cfg, logger, run_dir, False)
             except:
-                logger.exception("delete_work also failed")
+                logger.exception("delete_work failed")
             raise wrapper from exn
 
 
@@ -995,14 +1110,22 @@ def _eval_task_runtime(
                 )
             )
             memory_bytes = memory_max
-        ans["memory"] = memory_bytes
+        ans["memory_reservation"] = memory_bytes
+
+        memory_limit_multiplier = cfg["task_runtime"].get_float("memory_limit_multiplier")
+        if memory_limit_multiplier > 0.0:
+            ans["memory_limit"] = int(memory_limit_multiplier * memory_bytes)
 
     if "maxRetries" in runtime_values:
         ans["maxRetries"] = max(0, runtime_values["maxRetries"].coerce(Type.Int()).value)
+    if "preemptible" in runtime_values:
+        ans["preemptible"] = max(0, runtime_values["preemptible"].coerce(Type.Int()).value)
 
     if ans:
         logger.info(_("effective runtime", **ans))
-    unused_keys = list(key for key in runtime_values if key not in ans)
+    unused_keys = list(
+        key for key in runtime_values if key not in ("cpu", "memory") and key not in ans
+    )
     if unused_keys:
         logger.warning(_("ignored runtime settings", keys=unused_keys))
 
@@ -1015,37 +1138,75 @@ def _try_task(
     container: TaskContainer,
     command: str,
     runtime: Dict[str, Union[int, str]],
+    terminating: Callable[[], bool],
 ) -> None:
     """
-    Run the task command in the container, with up to runtime.maxRetries
+    Run the task command in the container, retrying up to runtime.preemptible occurrences of
+    Interrupted errors, plus up to runtime.maxRetries occurrences of any error.
     """
-    maxRetries = runtime.get("maxRetries", 0)
-    prevRetries = 0
+    max_retries = runtime.get("maxRetries", 0)
+    max_interruptions = runtime.get("preemptible", 0)
+    retries = 0
+    interruptions = 0
 
     while True:
+        if terminating():
+            raise Terminated()
         # copy input files, if needed
         if cfg["file_io"].get_bool("copy_input_files"):
             container.copy_input_files(logger)
 
         try:
             # start container & run command
-            return container.run(
-                logger, command, int(runtime.get("cpu", 0)), int(runtime.get("memory", 0))
-            )
-        except Exception as exn:
-            if isinstance(exn, Terminated) or prevRetries >= maxRetries:
-                raise
-            logger.error(
-                _(
-                    "task failure will be retried",
-                    error=exn.__class__.__name__,
-                    message=str(exn),
-                    prevRetries=prevRetries,
-                    maxRetries=maxRetries,
+            try:
+                return container.run(
+                    logger,
+                    command,
+                    int(runtime.get("cpu", 0)),
+                    int(runtime.get("memory_reservation", 0)),
+                    int(runtime.get("memory_limit", 0)),
                 )
+            finally:
+                if (
+                    "preemptible" in runtime
+                    and cfg.has_option("task_runtime", "_mock_interruptions")
+                    and interruptions < cfg["task_runtime"].get_int("_mock_interruptions")
+                ):
+                    raise Interrupted("mock interruption") from None
+                if terminating():
+                    raise Terminated()
+        except Exception as exn:
+            if isinstance(exn, Interrupted) and interruptions < max_interruptions:
+                logger.error(
+                    _(
+                        "interrupted task will be retried",
+                        error=exn.__class__.__name__,
+                        message=str(exn),
+                        prev_interruptions=interruptions,
+                        max_interruptions=max_interruptions,
+                    )
+                )
+                interruptions += 1
+            elif not isinstance(exn, Terminated) and retries < max_retries:
+                logger.error(
+                    _(
+                        "failed task will be retried",
+                        error=exn.__class__.__name__,
+                        message=str(exn),
+                        prev_retries=retries,
+                        max_retries=max_retries,
+                    )
+                )
+                retries += 1
+            else:
+                raise
+            container.reset(
+                logger,
+                interruptions + retries - 1,
+                delete_work=(
+                    cfg["file_io"]["delete_work"].strip().lower() in ["always", "failure"]
+                ),
             )
-            container.reset(logger, prevRetries)
-            prevRetries += 1
 
 
 def _eval_task_outputs(
@@ -1182,14 +1343,14 @@ def link_outputs(
 
 
 def _delete_work(cfg: config.Loader, logger: logging.Logger, run_dir: str, success: bool) -> None:
-    opt = cfg["task_runtime"]["delete_work"].strip().lower()
+    opt = cfg["file_io"]["delete_work"].strip().lower()
     if opt == "always" or (success and opt == "success") or (not success and opt == "failure"):
         if success and not cfg["file_io"].get_bool("output_hardlinks"):
             logger.warning(
-                "ignoring configuration [task_runtime] delete_work because it requires [file_io] output_hardlinks = true"
+                "ignoring configuration [file_io] delete_work because it requires also output_hardlinks = true"
             )
             return
-        for dn in ["write_", "work"]:
+        for dn in ["write_", "work", "failed_tries"]:
             dn = os.path.join(run_dir, dn)
             if os.path.isdir(dn):
                 shutil.rmtree(dn)
